# Simple support for organizing vcs branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

from bzrlib.commands import Command
from bzrlib.option import Option, ListOption, RegistryOption

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
import os
import tempfile

from bzrlib import (
    errors,
    osutils,
    trace,
    transport,
    urlutils,
    )

from bzrlib.plugins.scmproj import (
    action,
    project,
    )
""")


class cmd_project_init(Command):
    """Initialize new project meta-directory.

    For new local projects, this will make a new project dir with a .scmproj
    control branch inside.
    For new remote projects, this will only make a new scmproj control branch
    at the specified location.
    """
    aliases = ['pinit']
    takes_args = ['location']
    takes_options = []

    def run(self, location=None):
        project.Project.initialize(location)
        if project.is_url_local(location):
            if location == '.':
                directory = 'current directory'
            else:
                directory = '"%s"' % action.url_for_display(location)
            self.outf.write(('New ".scmproj" control branch intialized in %s.\n'
                'Edit %s in this directory to describe your project.\n')
                % (directory, project.CONFIG))
        else:
            self.outf.write(('New scmproj control branch intialized at "%s".\n'
                ) % action.url_for_display(location))


class cmd_project_info(Command):
    """Show info summary for project."""
    aliases = ['pinfo']
    takes_args = ['location?']
    takes_options = []

    def run(self, location=None):
        if location is None:
            location = '.'
        p = project.Project.open_containing(location)
        p.info(self.outf)


class cmd_project_get(Command):
    """Get a copy of an existing project on your local disk.

    This will make a new project dir with a .scmproj control branch inside,
    and/or get component branches of the project.

    FROM_LOCATION may refer to a local project dir or a remote scmproj
    control branch.
    If the TO_LOCATION is omitted, the last component of the FROM_LOCATION
    will be used.  In other words, "project-get ../my/proj" will attempt to
    create ./proj.

    To get both project control branch and component branches use command
    without --control-branch-only option::

        bzr project-get FROM_LOCATION [TO_LOCATION]

    To get only control branch of the project use command as::

        bzr project-get --control-branch-only FROM_LOCATION [TO_LOCATION]

    When you already have a local copy of a project with a control branch
    but without its components, and want to get component branches, then
    run this command from project root directory::

        bzr project-get --components-only

    Use other options (--alt, --subset) when necessary.
    """
    aliases = ['pget']
    takes_args = ['from_location?', 'to_location?']
    takes_options = [
        Option('control-branch-only',
            help="Only get the control branch (i.e. don't get the component branches)."),
        Option('components-only', help="Only get the component branches "
            "(when the control branch is already present)."),
        Option('alt', type=unicode, help='Get branches for the specified alt.'),
        Option('subset', type=unicode, help='Get components of the specified subset.'),
        Option('full-set', help='Get all components.'),
        ]

    def run(self, from_location=None, to_location=None, alt=None, subset=None,
        full_set=None, control_branch_only=None, components_only=None):
        if control_branch_only and components_only:
            raise errors.BzrError("You cannot use both --control-branch-only "
                "and --components-only options in the same time.")
        if not components_only:
            if from_location is None:
                raise errors.BzrError(
                    'You should specify source location as first argument.')
            # clone control branch first
            if to_location is None:
                to_location = urlutils.derive_to_location(from_location)
            project.Project(from_location).clone(to_location)
        else:
            # control branch already should be cloned
            if (from_location, to_location) != (None, None):
                raise errors.BzrError('To get only components branches '
                    'run project-get command without arguments.')
            to_location = '.'
        p = project.Project(to_location).open()
        if not control_branch_only:
            alt, component_list = p.cli_helper(alt=alt, subset=subset, full_set=full_set)
            p.update_workspace(self.outf, alt, component_list)
        p.save_settings(alt, subset)


class cmd_project_update(Command):
    """Update components of the project from their origins."""

    aliases = ['pup', 'pupdate']
    takes_args = ['location?']
    takes_options = [
        Option('components-only', help="Update only component branches."),
        Option('all', short_name='A', help='Force update all components.'),
        ]

    def run(self, location=None, components_only=None, all=False):
        if location is None:
            location = '.'
        p = project.Project.open_containing(location)
        if not components_only:
            p.update_control_branch(self.outf)
        alt, component_list = p.cli_helper()
        p.update_workspace(self.outf, alt, component_list, all)


class cmd_project_publish(Command):
    """Publish (push back) changes from components branches and control branch."""
    aliases = ['pback', 'project-push', 'ppush']
    takes_args = ['component*']
    takes_options = [
        Option('alt', argname='alt', type=unicode, help='Push data for specified alt.'),
        Option('subset', type=unicode, help='Push data for specified subset.'),
        Option('full-set', help='Push data for all components.'),
        Option('components-only', help="Publish only component branches."),
        Option('strict',
               help='Refuse to push if there are uncommitted changes in'
               ' components trees, --no-strict disables the check.'),
        Option('all', short_name='A', help='Force publish all components.'),
        ]

    def run(self, component_list=None, alt=None, subset=None, full_set=None,
        components_only=None, strict=None, all=False):
        p = project.Project.open_containing('.')
        alt, klist = p.cli_helper(alt=alt, subset=subset, full_set=full_set)
        if component_list is None:
            component_list = klist
        p.publish(self.outf, alt, component_list, components_only, strict, all)


class cmd_project_command(Command):
    """Run some command(s) for project components.

    .. note::
    
       Run this command from the root of local project copy.

    Option --cd-relpath (-C) will force execution of command(s) in the
    components working directories. It's equal to have "cd {RELPATH}"
    prefixed to each command. This is default behavior. If you need
    to suppress automatic directory change, please use --no-cd option.

    If you want to check the actual commands and its arguments for every
    component without actually invoking these commands, then you may run
    with options --dry-run and --verbose (-v), e.g::

        bzr pcmd -v --dry-run "missing {BRANCH_URL}"
    """
    _see_also = ['chdir']
    aliases = ['pcmd', 'prun', 'pcommand']
    takes_args = ['command+']
    takes_options = [
        'verbose',
        'dry-run',
        Option('ignore-errors', short_name='E',
            help="Don't stop on user errors."),
        Option('alt', argname='alt', type=unicode,
            help='Run command for specified alt.'),
        Option('subset', type=unicode,
            help='Run command for specified subset.'),
        Option('full-set', help='Run command for all components.'),
        Option('cd-relpath', short_name='C',
            help="Cd to component working directory before executing commands."),
        Option('shell', short_name='s',
            help='Run shell commands instead of bzr commands.'),
        Option('var1', type=unicode,
            help='Additional template variable {var1}.'),
        Option('var2', type=unicode,
            help='Additional template variable {var2}.'),
        Option('var3', type=unicode,
            help='Additional template variable {var3}.'),
        ListOption('include', short_name='i', type=unicode,
            help="Run command only for specified components."),
        ListOption('exclude', short_name='x', type=unicode,
            help="Don't run command for specified components."),
        RegistryOption('select',
            help="Run command only for components which conform the status.",
            value_switches=True,
            lazy_registry=('bzrlib.plugins.scmproj.vcs', 'checkers_reg'),
            ),
        ]

    def run(self, command_list, ignore_errors=False, alt=None, subset=None,
        full_set=None, cd_relpath=True, verbose=False, dry_run=False,
        shell=False, var1=None, var2=None, var3=None,
        include=None, exclude=None, changed=None, select=None):
        p = project.Project.open_containing('.')
        alt, component_list = p.cli_helper(alt=alt, subset=subset,
                                           full_set=full_set)
        if include and exclude:
            raise errors.BzrError("You can't use --include and --exclude "
                "in the same time.")
        if include:
            new_component_list = []
            for k in component_list:
                if k in include:
                    new_component_list.append(k)
            component_list = new_component_list
        if exclude:
            new_component_list = []
            for k in component_list:
                if k not in exclude:
                    new_component_list.append(k)
            component_list = new_component_list
        vars_dict = p.vars_helper(var1=var1, var2=var2, var3=var3)
        return p.run_cmds(command_list, ignore_errors=ignore_errors,
            to_file=self.outf, alt_name=alt, components=component_list,
            cd_relpath=cd_relpath, verbose=verbose, dry_run=dry_run,
            shell=shell, vars_dict=vars_dict, checker_func=select)


class cmd_chdir(Command):
    """Run bzr command in another directory.

    This command changes the current directory to another one and then
    launches the bzr command provided as the argument, e.g.::

        bzr chdir foo/bar -- missing http://example.com/spam

    This is equivalent to::

        # remember $CWD
        cd foo/bar
        bzr missing http://example.com/spam
        cd $ORIG_CWD

    This is ad-hoc solution to workaround some bugs or missing
    features in core bzr commands for batch processing.
    Intended to use with command `project-command`.

    .. note::
    
       You always should use ``--`` after directory you want to cd
       and before actual subcommand you want to execute in that directory.
       Otherwise it's not possible to pass options to the subcommand, e.g.::

        bzr chdir foo/bar -- missing --line http://example.com/spam

       This command will not work without ``--`` between ``foo/bar``
       and ``missing``.
    """
    aliases = ['cd']
    takes_args = ['directory', 'argv+']
    takes_options = []
    hidden = True

    def run(self, directory, argv_list):
        from bzrlib.commands import run_bzr
        cwd = os.getcwdu()
        os.chdir(directory)
        try:
            return run_bzr(argv_list)
        finally:
            os.chdir(cwd)


class cmd_project_commit(Command):
    """Commit changes in components working copies."""
    aliases = ['pci', 'pcommit']
    takes_args = ['component*']
    takes_options = [
        Option('subset', type=unicode, help='Commit components of specified subset.'),
        Option('full-set', help='Commit all components.'),
        Option('file', type=str, short_name='F', argname='msgfile',
               help='Take commit message from this file.'),
        Option('message', type=unicode, short_name='m',
               help="Commit message."),
        Option('components-only', help="Commit only changes in components."),
        ]

    def run(self, component_list=None, subset=None, full_set=None, file=None,
        message=None, components_only=None):
        if file is None and not message:
            raise errors.BzrError('You should provide commit message.')
        elif message and file:
            raise errors.BzrCommandError("please specify either --message or --file")
        p = project.Project.open_containing('.')
        if component_list is None:
            component_list = p.get_component_list(subset, full_set)
        if message:     # save commit message to temp file
            tmp_fileno, file = tempfile.mkstemp()
            msgfile = os.fdopen(tmp_fileno, 'w')
            try:
                msgfile.write(message.encode(osutils.get_user_encoding()))
            finally:
                msgfile.close()
        try:
            return p.commit(file, to_file=self.outf, component_list=component_list,
                components_only=components_only)
        finally:
            if message:     # if message have been saved we need to delete temp file
                try:
                    os.unlink(file)
                except IOError, e:
                    trace.warning("failed to unlink %s: %s; ignored", file, e)
