# Simple support for organizing bzr branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

# TODO: split it to several files and put them into tests/ subdir.

from cStringIO import StringIO
import os

from bzrlib import (
    branch,
    commands,
    errors,
    osutils,
    trace,
    urlutils,
    workingtree,
    )
from bzrlib.tests import TestCase, TestCaseWithTransport
from bzrlib.tests.test_sftp_transport import TestCaseWithSFTPServer
from bzrlib.transport import get_transport

from bzrlib.plugins.scmproj import (
    project,
    )
from bzrlib.plugins.scmproj.action import Action

from bzrlib.plugins.scmproj.project import (
    Component,
    Project,
    ProjectLocal,
    ProjectRemote,
    is_url_local,
    safe_path,
    )
from bzrlib.plugins.scmproj.config import (
    CONFIG_TMPL,
    Config,
    )
from bzrlib.plugins.scmproj.tests.whiteutils import reindentBlock

INSTALL_CMD = getattr(commands, "install_bzr_command_hooks", lambda: None)


class TestComponent(TestCase):

    def test_init(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual(None, ko.description)
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('.', ko.relpath)
        self.assertEqual(['trunk', 'bar'], ko.branches)
        # check optional values
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [component "foo"]
            DESCRIPTION = Mega component!
            RELPATH = .
            VCS = bzr
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual('Mega component!', ko.description)
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('.', ko.relpath)
        self.assertEqual(['trunk', 'bar'], ko.branches)
        # mixed config
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [PROJECT]
            COMPONENTS = foo bar

            [COMPONENTS]
            RELPATH = {COMPONENT}
            BRANCH_URL = http://example.com/{COMPONENT}/trunk

            [component bar]
            RELPATH = .
            BRANCHES = dev
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual(None, ko.description)
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('foo', ko.relpath)
        self.assertEqual(['trunk'], ko.branches)
        ko = Component(cfg, 'bar')
        self.assertEqual(None, ko.description)
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('.', ko.relpath)
        self.assertEqual(['dev'], ko.branches)

    def test_get_component_from_config(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [component "foo"]
            RELPATH = .
            BRANCHES =
            """))
        self.assertEqual('.',
            Component(cfg, 'foo').relpath)

    def test_get_branch(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            
            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('trunk')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('bar', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)
        ##
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)

        kb = ko.get_branch('bar')
        self.assertEqual('bzr', kb.vcs)

        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            
            [component "foo"]
            DESCRIPTION = Mega component!
            RELPATH = .
            VCS = bzr
            BRANCHES = trunk bar baz
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            
                [[branch "baz"]]
                VCS = bzr
                BRANCH_URL = https://bzr.host.org/project
                REVISION = 1234
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)

        kb = ko.get_branch('baz')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('https://bzr.host.org/project', kb.url)
        self.assertEqual('1234', kb.revision)

        # optional format for branches
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [component "foo"]
            VCS = bzr
            FORMAT = pack-0.92
            RELPATH = .
            BRANCHES = trunk bar baz
            
                [[branch "bar"]]
                VCS = bzr
                FORMAT = rich-root-pack
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            
                [[branch "baz"]]
                BRANCH_URL = https://host.org/project
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('trunk')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('pack-0.92', kb.format)

        kb = ko.get_branch('bar')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)

        kb = ko.get_branch('baz')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('pack-0.92', kb.format)

        # standalone branch (not in shared repo) as '.'
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [COMPONENTS]
            BRANCH_URL = http://example.com/{COMPONENT}
            
            [component "foo"]
            RELPATH = .
            BRANCHES = .
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/foo', kb.url)
        self.assertEqual(None, kb.revision)

        # component URL might be empty, but branch URL should not
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [component "foo"]
            RELPATH = .
            BRANCHES = spam eggs
            
                [[branch "eggs"]]
                REVISION = -1
            """))
        ko = Component(cfg, 'foo')

        self.assertRaises(errors.BzrError, ko.get_branch('spam')._get_url)
        self.assertRaises(errors.BzrError, ko.get_branch('eggs')._get_url)

        # mixed config
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [PROJECT]
            COMPONENTS = foo bar

            [COMPONENTS]
            RELPATH = {COMPONENT}
            BRANCH_URL = http://example.com/{COMPONENT}/trunk

            [component bar]
            RELPATH = .
            BRANCHES = dev
            """))
        ko = Component(cfg, 'foo')
        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/foo/trunk', kb.url)
        self.assertEqual(None, kb.revision)
        kb = ko.get_branch('trunk')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/foo/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        ko = Component(cfg, 'bar')
        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/bar/trunk', kb.url)
        self.assertEqual(None, kb.revision)
        kb = ko.get_branch('dev')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/bar/trunk', kb.url)
        self.assertEqual(None, kb.revision)

    def test_branch_format_from_components_section(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            FORMAT = rich-root-pack
            
            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('trunk')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('bar', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)

    def test_branch_vcs_from_components_section(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            VCS = bzr

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar

                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('trunk')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('bar', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)

    def test_branch_url_for_component(self):
        # BRANCH_URL template could be defined in the component section
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk dev release
            BRANCH_URL = http://example.com/{COMPONENT}/{BRANCH}

                [[branch "release"]]
                BRANCH_URL = http://example.com/releases/{COMPONENT}
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual('http://example.com/foo/trunk',
            ko.get_branch('DEFAULT').url)
        self.assertEqual('http://example.com/foo/trunk',
            ko.get_branch('trunk').url)
        self.assertEqual('http://example.com/foo/dev',
            ko.get_branch('dev').url)
        self.assertEqual('http://example.com/releases/foo',
            ko.get_branch('release').url)

    def test_component_and_branch_vcs(self):
        cfg = Config(reindentBlock("""\
            ### project.cfg v1

            [PROJECT]
            VCS = bzr

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk
            BRANCH_URL = http://example.com/{COMPONENT}/{BRANCH}
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('bzr', ko.get_branch('trunk').vcs)

    def test_branch_url_as_win32_path(self):
        cfg = Config(reindentBlock(
            r"""### project.cfg v1

            [PROJECT]
            VCS = bzr

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk
            BRANCH_URL = C:\work\test\branch
            """))
        self.assertEqual("C:\\work\\test\\branch",
            Component(cfg, 'foo').get_branch('trunk').url)


class Test42(TestCase):

    def test_is_url_local(self):
        self.assertTrue(is_url_local('.'))
        self.assertTrue(is_url_local(urlutils.local_path_to_url('.')))
        self.assertFalse(is_url_local('http://example.com'))

    def test_project_constructor(self):
        self.assertTrue(isinstance(Project('.'), ProjectLocal))
        self.assertTrue(isinstance(Project('http://example.com'),
            ProjectRemote))

    def test_safe_path(self):
        self.assertEqual('foo', safe_path('foo'))
        self.assertEqual('_foo', safe_path('Foo'))
        self.assertEqual('__foo', safe_path('_foo'))
        self.assertEqual('foo~2fbar', safe_path('foo/bar'))
        self.assertEqual('foo~5cbar', safe_path('foo\\bar'))
        self.assertEqual('foo~7e', safe_path('foo~'))
        self.assertEqual('foo~2e', safe_path('foo.'))
        self.assertEqual('nu~6c.foo', safe_path('nul.foo'))
        self.assertEqual('com~31.foo', safe_path('com1.foo'))


class TestProject(TestCaseWithTransport):

    def setUp(self):
        TestCaseWithTransport.setUp(self)
        INSTALL_CMD()

    def test__create_skeleton(self):
        Project._create_skeleton('.')
        t = get_transport('.')
        self.assertIsDirectory('.scmproj', t)
        self.assertIsDirectory('.scmproj/.bzr', t)
        self.assertIsDirectory('.scmproj/alts', t)
        self.failUnlessExists([
            '.scmproj/project.cfg',
            '.scmproj/readme.txt',
            ])
        # check we have valid branch there
        br = branch.Branch.open('.scmproj')
        # check the revision tree has project config in it
        br.lock_read()
        try:
            rt = br.repository.revision_tree(br.last_revision())
            self.assertNotEqual(None, rt.path2id('project.cfg'))
        finally:
            br.unlock()

    def test__check_control_branch(self):
        wt = self.make_branch_and_tree('.')
        self.build_tree(['project.cfg'])
        wt.add('project.cfg')
        # no committed project config
        self.assertRaises(errors.BzrError,
            Project._check_control_branch, wt.branch)
        #
        wt.commit('commit and check')
        Project._check_control_branch(wt.branch)


class TestProjectLocal(TestCaseWithTransport):

    def setUp(self):
        TestCaseWithTransport.setUp(self)
        INSTALL_CMD()

    def test_initialize(self):
        t = get_transport('.')
        Project.initialize(t.base)
        self.assertIsDirectory('.scmproj', t)
        self.assertIsDirectory('.scmproj/.bzr', t)
        self.assertIsDirectory('.scmproj/alts', t)
        self.failUnlessExists(['.scmproj/project.cfg'])
        # check we have valid branch there
        branch.Branch.open('.scmproj')
        # you can't create twice
        self.assertRaises(errors.BzrError, Project.initialize, t.base)

    def test_error_wo_open(self):
        Project.initialize('.')
        p = Project('.')
        self.assertRaises(errors.BzrError, p.get_alt_name)
        self.assertRaises(errors.BzrError, p.get_subset_name)

    def test_info_smoke(self):
        Project.initialize('.')
        p2 = Project('.').open()
        sio = StringIO()
        p2.info(sio)
        self.assertNotEqual('', sio.getvalue())

    def test_clone(self):
        self.build_tree(['a/', 'b/'])
        Project.initialize('a')

        # existing project dir
        Project('a').clone('b')
        p2 = Project('b').open()

        # non-existing project dir
        Project('a').clone('c')
        p2 = Project('c').open()

    def test_clone_simple_branch(self):
        # we cannot clone if branch has no project.cfg
        self.make_branch('.scmproj')
        self.assertRaises(errors.BzrError, Project('.').clone, 'b')

    def test_run_actions(self):
        Project.initialize('.')

        f = open(project.CONFIG_FN, 'wb')
        try:
            f.write(reindentBlock("""\
                ### project.cfg v1

                [PROJECT]
                VCS = bzr
                
                [COMPONENTS]
                BRANCH_URL = http://example.com/{BRANCH}
                
                [ALTS]
                dev = foo:bar
                test = foo:baz
                
                [component "foo"]
                RELPATH = .
                BRANCHES = trunk bar baz
                
                    [[branch "bar"]]
                    BRANCH_URL = http://another.net/branch
                    REVISION = xyz
                
                    [[branch "baz"]]
                    BRANCH_URL = https://host.org/project
                    REVISION = 1234
                
                [component "spam"]
                RELPATH = subdir
                BRANCHES = trunk
                    [[branch "trunk"]]
                    BRANCH_URL = bzr://server/branches/{BRANCH}
                """))
        finally:
            f.close()

        # actions for testing component iterator
        class LoggerAction(Action):

            def __call__(self, to_file, ko, kb, local_paths, local_urls):
                to_file.write(' ko.name = %s\n' % ko.name)
                to_file.write('  kb.name = %s' % kb.name)
                to_file.write('; kb.url = %s' % kb.url)
                to_file.write('\n')

        class DozingAction(Action):

            def __call__(self, to_file, ko, kb, local_paths, local_urls):
                to_file.write(' Z-z-z\n')

        action_list = [LoggerAction(), DozingAction()]

        # actually testing
        p2 = Project('.').open()

        # test simple
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio)
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = trunk; kb.url = http://example.com/trunk
             Z-z-z
            #------------------------------------------------------------
            # Component: spam
             ko.name = spam
              kb.name = trunk; kb.url = bzr://server/branches/trunk
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

        # test alt_name
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio,
            alt_name='dev')
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = bar; kb.url = http://another.net/branch
             Z-z-z
            #------------------------------------------------------------
            # Component: spam
             ko.name = spam
              kb.name = trunk; kb.url = bzr://server/branches/trunk
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

        # test component list
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio,
            alt_name='test', component_list=['foo'])
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = baz; kb.url = https://host.org/project
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

        # test cd_relpath
        os.mkdir('subdir')
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio, cd_relpath=True)
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = trunk; kb.url = http://example.com/trunk
             Z-z-z
            #------------------------------------------------------------
            # Component: spam
            #  Entering directory `subdir'
             ko.name = spam
              kb.name = trunk; kb.url = bzr://server/branches/trunk
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

    def test_get_alt_name(self):
        Project.initialize('.')
        p = Project('.').open()
        self.assertEqual('DEFAULT', p.get_alt_name())
        p.settings.alt = 'foo'
        self.assertEqual('foo', p.get_alt_name())

    def test_get_subset_name(self):
        Project.initialize('.')
        p = Project('.').open()
        self.assertEqual('DEFAULT', p.get_subset_name())
        p.settings.subset = 'bar'
        self.assertEqual('bar', p.get_subset_name())

    def test_save_settings(self):
        Project.initialize('.')

        p = Project('.').open()
        self.assertEqual('DEFAULT', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings(None, None)
        self.assertEqual('DEFAULT', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings('foo', None)
        self.assertEqual('foo', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings('boo', None)
        self.assertEqual('foo', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings('boo', None, override=True)
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings(None, 'spam')
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('spam', p.get_subset_name())

        p.save_settings(None, 'eggs')
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('spam', p.get_subset_name())

        p.save_settings(None, 'eggs', override=True)
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('eggs', p.get_subset_name())


class TestProjectRemote(TestCaseWithSFTPServer):

    def setUp(self):
        TestCaseWithSFTPServer.setUp(self)
        INSTALL_CMD()

    def test_initialize(self):
        t_local = get_transport('server')
        sftp_url = self.get_url('server')
        Project.initialize(sftp_url)
        self.assertIsDirectory('.bzr', t_local)
        # check we have valid branch there as remote
        branch.Branch.open(sftp_url)
        # check we have valid branch there as local
        branch.Branch.open('server')
        # you can't create twice
        self.assertRaises(errors.BzrError, Project.initialize, sftp_url)

    def test_initialize_to_exisiting_dir(self):
        t_local = get_transport('.')
        Project.initialize(self.get_url())
        self.assertIsDirectory('.bzr', t_local)
        # check we have valid branch there as remote
        branch.Branch.open(self.get_url())
        # check we have valid branch there as local
        branch.Branch.open('.')
        # you can't create twice
        self.assertRaises(errors.BzrError, Project.initialize, self.get_url())

    def test_clone(self):
        self.build_tree(['a/', 'b/'])
        Project.initialize(self.get_url('a'))
        Project(self.get_url('a')).clone('b')
        p2 = Project('b').open()

    def test_clone_simple_branch(self):
        # we cannot clone if branch has no project.cfg
        self.make_branch('server')
        self.assertRaises(errors.BzrError, Project(self.get_url('server')).clone, 'b')

    def test_open(self):
        p = Project(self.get_url('.'))
        self.assertRaises(errors.BzrError, p.open)


class TestSubprojects(TestCaseWithTransport):

    def setUp(self):
        TestCaseWithTransport.setUp(self)
        INSTALL_CMD()

    def _prepare(self):
        # prepare testing objects
        # subproject
        foo = self.make_branch_and_tree('foo')
        self.build_tree(['foo/dir/', 'foo/file'])
        foo.add(['dir', 'file'])
        foo.commit('initial revision in foo')
        #
        Project.initialize('subproject')
        subproject = workingtree.WorkingTree.open('subproject/.scmproj')
        self.build_tree_contents([
            (os.path.join('subproject', project.CONFIG_FN),
            reindentBlock("""\
                ### project.cfg v1

                [component "foo"]
                RELPATH = .
                BRANCH_URL = %s
                """) % urlutils.local_path_to_url(os.path.abspath('foo'))),
            ])
        subproject.commit('config of subproject')
        # main project
        bar = self.make_branch_and_tree('bar')
        self.build_tree(['bar/ddd/', 'bar/fff'])
        bar.add(['ddd', 'fff'])
        bar.commit('initial revision in bar')
        #
        Project.initialize('project')
        main_project = workingtree.WorkingTree.open('project/.scmproj')
        self.build_tree_contents([
            (os.path.join('project', project.CONFIG_FN),
             reindentBlock("""\
                ### project.cfg v1

                [component "bar"]
                RELPATH = .
                BRANCH_URL = %(bar)s

                [subproject "spam"]
                RELPATH = spam
                SUBPROJECT_URL = %(spam)s
                """) % {
                    'bar': urlutils.local_path_to_url(os.path.abspath('bar')),
                    'spam': urlutils.local_path_to_url(os.path.abspath('subproject')),
                    }),
            ])
        main_project.commit('config of main project')

    def test_smoke(self):
        self._prepare()
        Project('subproject').open()
        Project('project').open()

    def test_clone(self):
        self._prepare()
        Project('project').clone('aaa')
        t = get_transport('.')
        self.assertIsDirectory('aaa/.scmproj', t)
        self.failUnlessExists(['aaa/.scmproj/project.cfg'])

    def test_create_components_and_subprojects(self):
        self._prepare()
        Project('project').clone('aaa')
        p = Project('aaa').open()
        p.update_workspace()    # it should clone all subprojects too
        t = get_transport('.')
        self.assertIsDirectory('aaa/ddd', t)
        self.failUnlessExists(['aaa/fff'])
        self.assertIsDirectory('aaa/spam/', t)
        self.assertIsDirectory('aaa/spam/.scmproj', t)
        self.failUnlessExists(['aaa/spam/.scmproj/project.cfg'])
        self.assertIsDirectory('aaa/spam/dir', t)
        self.failUnlessExists(['aaa/spam/file'])

    def test_run_actions(self):
        self._prepare()
        Project('project').clone('aaa')
        p = Project('aaa').open()
        p.update_workspace()

        # actions for testing component iterator
        class LoggerAction(Action):

            def __call__(self, to_file, ko, kb, local_paths, local_urls):
                to_file.write(' ko.name = %s\n' % ko.name)
                #to_file.write('  kb.name = %s' % kb.name)
                #to_file.write('; kb.url = %s' % kb.url)
                #to_file.write('\n')

        class DozingAction(Action):

            def __call__(self, to_file, ko, kb, local_paths, local_urls):
                to_file.write(' Z-z-z\n')

        action_list = [LoggerAction(), DozingAction()]

        # actually testing
        p2 = Project('aaa').open()

        # test simple
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio, cd_relpath=True)
        self.assertEquals(reindentBlock("""\
            # Component: bar
             ko.name = bar
             Z-z-z
            #------------------------------------------------------------
            # Subproject: spam
            #  Entering directory `spam'
            #------------------------------------------------------------
            # Component: foo
             ko.name = foo
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

    def test_info(self):
        self._prepare()
        Project('project').clone('aaa')
        p = Project('aaa').open()
        p.update_workspace()

        sio = StringIO()
        p2 = Project('aaa').open()
        p2.info(sio)
        result = sio.getvalue()
        # strip out cwd path to make testing better
        cwd = osutils.getcwd()
        if not cwd.endswith('/'):
            cwd = cwd + '/'
        url = urlutils.local_path_to_url(cwd)
        if not url.endswith('/'):
            url = url + '/'
        result = result.replace(url, '$URL/')
        result = result.replace(cwd, '$DIR/')
        self.assertEqual(reindentBlock("""\
            Project: unnamed
            URL: $DIR/aaa/
            Parent URL: $DIR/project/.scmproj/

            Components:
              [bar]
                  relpath = .
                  branch  = $URL/bar

            Subprojects:
              [spam]
                  Project: unnamed
                  URL: $DIR/aaa/spam/
                  Parent URL: $DIR/subproject/.scmproj/

                  Components:
                    [foo]
                        relpath = .
                        branch  = $URL/foo
            """), result)
